from umongo import Document, Schema, fields
from .database import db


class Meeting(Document):

    class Schema(Schema):
        calendar = fields.ReferenceField("Calendar", required=True)
        name = fields.StringField(max_length=120, required=True)
        users = fields.ListField(fields.EmailField(max_length=255), required=True)
        begin = fields.DateTimeField(required=True)
        duration = fields.IntField(required=True)
        place = fields.StringField(max_length=255)
        tags = fields.ListField(fields.StringField(max_length=20))
        comment = fields.StringField(max_length=255)

    class Config:
        collection = db.meeting



from datetime import datetime

from ..model.meeting import Meeting
from ..view import create_calendar
from ..view.meeting_api import (
    create_meeting, list_meetings, get_meeting, modify_meeting, event_meeting_creation)

from .common import BaseTest


def calendar_factory(name="test", users=["test@test.com"]):
    r = create_calendar("test", ["test@test.com"])
    return r


class TestMeetingAPI(BaseTest):

    def test_create_meeting(self):
        calendar = calendar_factory()
        r = create_meeting(calendar_id=calendar.pk, name="first meeting", users=["pika@chu.com"],
                           duration="90", place="ici et la bas", comment="pourquoi pas", tags=['pokemon'])
        assert r
        meeting = Meeting.find_one(r.pk)
        assert meeting.pk == r.pk

    def test_list_meeting(self):
        calendar = calendar_factory()
        r = create_meeting(calendar_id=calendar.pk, name="meeting1")
        r = create_meeting(calendar_id=calendar.pk, name="meeting2")
        r = create_meeting(calendar_id=calendar.pk, name="meeting3")
        r = create_meeting(calendar_id=calendar.pk, name="meeting4")
        r = create_meeting(calendar_id=calendar.pk, name="meeting5")
        r = list_meetings(calendar.pk)
        assert r
        assert r.count() == 5

    def test_get_meeting(self):
        calendar = calendar_factory()
        r = create_meeting(calendar_id=calendar.pk, name="meeting1")
        r = get_meeting(r.pk)
        assert r.data.name == "meeting1"

    def test_modify_meeting(self):
        calendar = calendar_factory()
        r = create_meeting(calendar_id=calendar.pk, name="meeting1")
        r = get_meeting(r.pk)
        assert r
        assert r.data.name == "meeting1"
        r = modify_meeting(r.pk, name="new_name")
        assert r
        assert r.data.name == "new_name"
        r = modify_meeting(r.pk, users=["pika@chu.com"])
        assert r
        assert len(r.data.users) == 1
        assert r.data.users[0] == "pika@chu.com"
        r = modify_meeting(r.pk, users=["tor@chu.com"], append=True)
        assert r
        assert len(r.data.users) == 2
        assert r.data.users[1] == "tor@chu.com"
        r = modify_meeting(r.pk, date_begin=datetime(2007, 12, 5, 12, 00))
        assert r
        assert r.data.begin == datetime(2007, 12, 5, 12, 00)
        r = modify_meeting(r.pk, duration=100)
        assert r
        assert r.data.duration == 100
        r = modify_meeting(r.pk, place="defined")
        assert r
        assert r.data.place == "defined"



import jwt
from passlib.apps import custom_app_context as pwd_context

from .config import SECRET_KEY


def encrypt_password(password):
    return pwd_context.encrypt(password)


def verify_password(password, pwd_hash):
    return pwd_context.verify(password, pwd_hash)


def encode_token(payload):
    return jwt.encode(payload, SECRET_KEY, algorithm='HS256').decode()


def decode_token(token):
    try:
        return jwt.decode(token, SECRET_KEY)
    except jwt.InvalidTokenError:
        return None



import mongoengine


def abort(*args, **kwargs):
    raise RuntimeError(args, kwargs)


class BaseController:

    """
    Controller base class, providing usefull function for handling document
    """

    def __init__(self, document):
        self.document = document

    def save_or_abort(self, abort=abort, if_match=None):
        try:
            if if_match is True:
                self.document.save(save_condition={"doc_version": self.document.doc_version})
            elif if_match:
                self.document.save(save_condition={"doc_version": if_match})
            else:
                self.document.save()
        except mongoengine.ValidationError as e:
            errors = e.to_dict()
            if errors:
                # ValidationErrors issued in the clean function are wrapped
                # in a useless NON_FIELD_ERRORS
                non_field_errors = errors.pop(mongoengine.base.NON_FIELD_ERRORS, {})
                if isinstance(non_field_errors, dict):
                    errors.update(non_field_errors)
                    abort(400, **errors)
                else:
                    abort(400, non_field_errors)
            else:
                abort(400, e.message)
        except mongoengine.errors.NotUniqueError as e:
            abort(400, str(e))
        except mongoengine.errors.FieldDoesNotExist as e:
            abort(400, str(e))

    def update(self, payload):
        for key, value in payload.items():
            setattr(self.document, key, value)


class ControlledDocument(mongoengine.Document):

    """
    Mongoengine abstract document providing a controller attribute to
    alter with style the document !
    """
    meta = {'abstract': True, 'controller_cls': BaseController}

    @property
    def controller(self):
        controller_cls = self._meta.get('controller_cls')
        if not controller_cls:
            raise NotImplementedError('No controller setted for this document')
        return controller_cls(self)

    def clean(self):
        """Automatically called at save time, triggers controller's clean"""
        ctrl = self.controller
        if hasattr(ctrl, 'clean'):
            ctrl.clean()



from autobahn.twisted.wamp import Application

app = Application("butler")

app._start = {}
app._waiting = {}


def _start_app(application_name, version):
    global app
    app._start[application_name] = version
    app.session.publish('butler.' + application_name + '.start')
    for key in app._waiting:
        if key not in app._start:
            application = app._waiting[key]
            _try_to_start_app(
                application['name'], application['version'], application['required_components'])


def _add_to_waiting_list(application_name, version, required_components):
    global app

    if application_name in app._waiting:
        del app._waiting[application_name]
    app._waiting[application_name] = {
        'name': application_name, 'version': version, 'required_components': required_components}


def _try_to_start_app(application_name, version, required_components):
    global app
    if not required_components:
        _start_app(application_name, version)
    else:
        start_component = []
        for key in required_components.keys():
            if key in app._start and required_components[key] == app._start[key]:
                start_component.append(key)
        if len(start_component) == len(required_components):
            _start_app(application_name, version)
        else:
            _add_to_waiting_list(application_name, version, required_components)


@app.register()
def register(application_name, version, required_components=None):
    """
    application_name : name of your application
    version : version of your application
    required_components dictionary of components required for you application
    and their version required

        {
           "component" : "1.1",
           "component2" : "0.1",
           ...
        }

     the butler will publish an event:

         butler.yourapplicationname.start

     when all the different component required has been register
     """
    global app
    _try_to_start_app(application_name, version, required_components)
    for k in app._start:
        if k in app._waiting:
            del app._waiting[k]


if __name__ == '__main__':
    app.run(url="ws://127.0.0.1:8080/ws")



from ..model import Calendar

event_calendar_creation = "calendar.created"


def create_calendar(name, users=None):
    if not users:
        # todo call auth module to get the current user
        users = ['test@test.com']
    calendar = Calendar(name=name, users=users)
    calendar.commit()
    return calendar


def list_calendar(page=1, per_page=20):
    skip = (page - 1) * per_page
    calendars = Calendar.find(skip=skip, limit=per_page)
    return calendars


def get_calendar(calendar_id):
    calendar = Calendar.find_one(calendar_id)
    return calendar


def modify_calendar(calendar_id, name=None, users=None, append=False):
    errors = {}
    if not isinstance(append, bool):
        errors['append'] = 'Must be boolean'
    if name is not None and (not isinstance(name, str) or len(name) > 255):
        errors['name'] = 'Must be string or empty'
    if users is not None and not isinstance(users, list):
        errors['users'] = 'Must be list or empty'

    if users:
        if [u for u in users if not isinstance(u, str) or len(u) > 255]:
            errors['users'] = 'List element must be strings'

    if errors:
        return None

    calendar = Calendar.find_one(calendar_id)
    if not calendar:
        return None
    if name:
        calendar.data.name = name
    if users:
        if append:
            calendar.data.users.extend(users)
        else:
            calendar.data.users = users
    calendar.commit()
    return calendar


def remove_user(calendar_id, user):
    calendar = Calendar.find_one(calendar_id)
    if not calendar:
        return None
    if user in calendar.data.users:
        calendar.data.users.remove(user)
        calendar.commit()
    return calendar


def add_user(calendar_id, user):
    calendar = Calendar.find_one(calendar_id)
    if not calendar:
        return None
    if user in calendar.data.users:
        return calendar
    calendar.data.users.append(user)
    calendar.commit()
    return calendar



from mongoengine import Document, DynamicDocument
from xin.bb.model_util.model import BaseDocument, Marshallable
from xin.bb.model_util.controller import ControlledDocument, BaseController
from xin.bb.model_util.searcher import BaseSolrSearcher, Searcher, SearchableDocument
from xin.bb.model_util.version import VersionedDocument, HistorizedDocument
from xin.bb.model_util import fields


__all__ = ('Document', 'DynamicDocument',
           'BaseDocument', 'ControlledDocument', 'BaseController', 'fields',
           'BaseSolrSearcher', 'Searcher', 'SearchableDocument', 'Marshallable',
           'HistorizedDocument', 'VersionedDocument')



#! /usr/bin/env python3

from .view import (
    create_calendar, list_calendar, get_calendar, modify_calendar, add_user, remove_user,
    create_meeting, list_meetings, get_meeting, modify_meeting, delete_meeting)

from bson import ObjectId
import autobahn_sync


def treat_data(data):
    if not data:
        return 404, {'message', 'Not Found'}
    return 200, data.dump()


@autobahn_sync.register('kalendar.create')
def calendar_create(name, users):
    calendar = create_calendar(name, users)
    return treat_data(calendar)


@autobahn_sync.register('kalendar.list')
def calendar_list(page=1, per_page=20):
    cursor = list_calendar(page, per_page)
    calendars = [c.dump() for c in cursor]
    return 200, calendars


@autobahn_sync.register('kalendar.get')
def calendar_get(calendar_id):
    calendar = get_calendar(ObjectId(calendar_id))
    return treat_data(calendar)


@autobahn_sync.register('kalendar.modify')
def calendar_modify(calendar_id, name=None, users=None, append=False):
    calendar = modify_calendar(ObjectId(calendar_id), name, users, append)
    return treat_data(calendar)


@autobahn_sync.register('kalendar.add_user')
def calendar_add_user(calendar_id, user):
    calendar = add_user(ObjectId(calendar_id), user)
    return treat_data(calendar)


@autobahn_sync.register('kalendar.remove_user')
def calendar_remove_user(calendar_id, user):
    calendar = remove_user(ObjectId(calendar_id), user)
    return treat_data(calendar)


@autobahn_sync.register('kalendar.meeting.create')
def meeting_create(calendar_id, name, users, date_begin, duration=None,
                   place=None, comment=None, tags=None):
    meeting = create_meeting(ObjectId(calendar_id), name, users, date_begin,
                             duration, place, comment, tags)
    return treat_data(meeting)


@autobahn_sync.register('kalendar.meeting.list')
def meeting_list(calendar_id):
    return list_meetings(calendar_id)


@autobahn_sync.register('kalendar.meeting.get')
def meeting_get(meeting_id):
    meeting = get_meeting(ObjectId(meeting_id))
    return treat_data(meeting)


@autobahn_sync.register('kalendar.meeting.modify')
def meeting_modify(meeting_id, name=None, users=None, append=False, date_begin=None,
                   duration=None, place=None, comment=None, tags=None):
    meeting = modify_meeting(meeting_id, name, users, append, date_begin,
                             duration, place, comment, tags)
    return treat_data(meeting)


@autobahn_sync.register('kalendar.meeting.delete')
def meeting_delete(meeting_id):
    if delete_meeting(meeting_id):
        return 204
    return 400, {'message': 'do not succeed to delete the meeting'}


@autobahn_sync.on_challenge
def on_challenge(challenge):
    import json
    return json.dumps({'application_name': 'kalendar', 'version': 0.1})


def main():
    autobahn_sync.run(blocking=True, authid="app", authmethods=[u"ticket"])


if __name__ == '__main__':
    main()



from autobahn.wamp.exception import ApplicationError
from autobahn.twisted.wamp import ApplicationSession, Application
from twisted.internet import reactor
from twisted.web.server import Site
from twisted.internet.defer import inlineCallbacks, returnValue
from datetime import datetime

from .config import LISTEN_PORT, RETRIEVE_USER_RPC
from .rest import rest_api_factory
from .tools import decode_token


class AuthenticSession(ApplicationSession):

    def onJoin(self, details):

        rest_app = rest_api_factory(self)
        reactor.listenTCP(LISTEN_PORT, Site(rest_app.resource()))

        @inlineCallbacks
        def authenticate(realm, login, details):
            token = decode_token(details['ticket'])
            if (not token or login != token['login'] or
                    token['exp'] < datetime.utcnow().timestamp()):
                raise ApplicationError("xin.authentic.invalid_ticket",
                    "could not authenticate session - invalid token"
                    " '{}' for user {}".format(token, login))
                raise ApplicationError('Invalid token')
            user = yield self.call(RETRIEVE_USER_RPC, login)
            if not user:
                raise ApplicationError("xin.authentic.no_such_user",
                    "could not authenticate session - no such user {}".format(login))
                raise ApplicationError('Unknown user')
            print("[AUTHENTIC] WAMP-Ticket dynamic authenticator invoked: realm='{}', "
                  "authid='{}', ticket='{}'".format(realm, login, details))
            returnValue(user.get('role'))

        return self.register(authenticate, 'xin.authentic.authenticate')


if __name__ == "__main__":
    import sys
    from twisted.python import log
    log.startLogging(sys.stdout)

    wampapp = Application()
    wampapp.run("ws://localhost:8080", "realm1", standalone=False)



from mongoengine import Document

from xin.bb.model_util.version import HistorizedDocument
from xin.bb.model_util.searcher import SearchableDocument
from xin.bb.model_util.controller import ControlledDocument


class Marshallable(Document):

    """
    Special unmarshal to properly handle properties
    """
    meta = {'abstract': True}

    class Marshaller:

        def __init__(self, doc):
            self.doc = doc

        def __getitem__(self, key):
            return getattr(self.doc, key)

    def __marshallable__(self):
        return self.Marshaller(self)


class BaseDocument(ControlledDocument, Marshallable, HistorizedDocument,
                   SearchableDocument):

    """
    Document default class, all actual documents should inherit from this one
    """
    meta = {'abstract': True}



from flask import url_for
from collections import namedtuple
from bson import ObjectId

from xin.bb.tools import get_pagination_urlargs, abort
from xin.bb.core_resource import CoreResource
from xin.bb import view_util
from mongoengine.fields import SequenceField, IntField, ObjectIdField


HistoryAPI = namedtuple('HistoryAPI', ('item', 'endpoint', 'list',
                                       'endpoint_list', 'schema'))


def convert_id(obj_id_cls, obj_id):
    try:
        if isinstance(obj_id_cls, (SequenceField, IntField)):
            return int(obj_id)
        elif isinstance(obj_id_cls, (ObjectIdField)):
            return ObjectId(obj_id)
        else:
            return obj_id
    except ValueError:
        abort(400, origin='Invalid origin value type')


def history_api_factory(origin_cls):
    if 'history_cls' not in origin_cls._meta:
        raise RuntimeError('%s must have a `history_cls` meta field' % origin_cls)
    history_cls = origin_cls._bootstrap_history_cls()

    def history_link_builder(obj):
        return {'self': url_for('HistoryAPI', origin_id=obj.origin.id,
                                item_id=obj.id)}

    endpoint = "%sHistoryAPI" % origin_cls.__name__
    endpoint_list = "%sHistoryListAPI" % origin_cls.__name__
    endpoint_origin = "%sAPI" % origin_cls.__name__

    class HistorySchema(view_util.UnknownCheckedSchema):

        """Read-only schema"""

        def get_links(self, obj):
            return {
                'self': url_for(endpoint, origin_id=obj.origin.id,
                                item_id=obj.id),
                'parent': url_for(endpoint_list, origin_id=obj.origin.id),
                'origin': url_for(endpoint_origin, item_id=obj.origin.id)
            }

        class Meta:
            model = history_cls

    class HistoryItemAPI(CoreResource):

        def get(self, origin_id, item_id):
            item = history_cls.objects.get_or_404(id=item_id)
            if str(item.origin.id) != origin_id:
                abort(404)
            return HistorySchema().dump(item).data

    class HistoryListAPI(CoreResource):

        def get(self, origin_id):
            page, per_page = get_pagination_urlargs()
            origin_id = convert_id(origin_cls.id, origin_id)
            items = history_cls.objects(
                origin=origin_id).paginate(page=page, per_page=per_page)
            links = {'origin': url_for(endpoint_origin, item_id=origin_id)}
            route = url_for(endpoint_list, origin_id=origin_id)
            return view_util.PaginationSerializer(HistorySchema(), route).dump(
                items, links=links).data

    return HistoryAPI(HistoryItemAPI, endpoint, HistoryListAPI,
                      endpoint_list, HistorySchema)


def register_history(api, origin_cls, name=None):
    name = name or origin_cls.__name__.lower()
    history = history_api_factory(origin_cls)
    api.add_resource(history.list, '/%s/<string:origin_id>/history' % name,
                     endpoint=history.endpoint_list)
    api.add_resource(history.item,
                     '/%s/<string:origin_id>/history/<objectid:item_id>' % name,
                     endpoint=history.endpoint)



from .calendar_api import (create_calendar, list_calendar, get_calendar,
                           modify_calendar, event_calendar_creation, add_user, remove_user)
from .meeting_api import (create_meeting, list_meetings, get_meeting,
                          modify_meeting, delete_meeting, event_meeting_creation)

__all__ = ('create_calendar', 'list_calendar', 'get_calendar',
           'modify_calendar', 'event_calendar_creation', 'add_user', 'remove_user',
           'create_meeting', 'list_meetings', 'get_meeting',
           'modify_meeting', 'event_meeting_creation', 'delete_meeting')



import json
from datetime import datetime
from klein import Klein
from twisted.internet.defer import inlineCallbacks, returnValue

from .tools import encrypt_password, verify_password, encode_token
from .config import RETRIEVE_USER_RPC, REGISTER_USER_RPC, TOKEN_VALIDITY, CORS_ORIGIN


class Response400(Exception):
    pass


def _retreive_content(request):
    try:
        body = json.loads(request.content.read().decode())
    except ValueError:
        raise Response400('Invalid JSON body.')
    login = body.get('login')
    password = body.get('password')
    if type(login) is not str or type(password) is not str:
        raise Response400('Missing or invalid login and/or password.')
    return login, password, body


def _set_cors_headers(request):
    request.setHeader(b'Access-Control-Allow-Origin', CORS_ORIGIN)
    request.setHeader(b'Access-Control-Allow-Headers', b'accept, content-type')
    request.setHeader(b'Access-Control-Allow-Method', b'POST, OPTIONS')


def _create_token(login):
    exp = datetime.utcnow().timestamp() + TOKEN_VALIDITY
    return encode_token({'login': login, 'exp': exp})


def rest_api_factory(wamp_session):

    klein_app = Klein()

    @klein_app.handle_errors(Response400)
    def response_400(request, failure):
        request.setResponseCode(400)
        return json.dumps({'message': str(failure.value)})

    @klein_app.route('/login', methods=['POST', 'OPTIONS'])
    @inlineCallbacks
    def login(request):
        _set_cors_headers(request)
        if request.method == b'OPTIONS':
            return
        login, password, body = _retreive_content(request)
        user = yield wamp_session.call(RETRIEVE_USER_RPC, login)
        if (not user or not user.get('hashed_password') or
                not verify_password(password, user['hashed_password'])):
            raise Response400('Unknown user or invalid password')
        returnValue(json.dumps({'login': login, 'token': _create_token(login)}))

    @klein_app.route('/signin', methods=['POST', 'OPTIONS'])
    @inlineCallbacks
    def signin(request):
        _set_cors_headers(request)
        if request.method == b'OPTIONS':
            return
        login, password, body = _retreive_content(request)
        hashed_password = encrypt_password(password)
        user = yield wamp_session.call(REGISTER_USER_RPC, login, hashed_password, body)
        if not user:
            raise Response400("Couldn't create user")
        returnValue(json.dumps({'login': login, 'token': _create_token(login)}))

    # @klein_app.route('/login/github')
    # @inlineCallbacks
    # def login_github(request):
    #   pass

    return klein_app



from autobahn_sync.core import AutobahnSync, DEFAULT_AUTOBAHN_ROUTER, DEFAULT_AUTOBAHN_REALM


class OnJoined:

    def __init__(self, application_name, version, required_components, app, callback=None):
        self.application_name = application_name
        self.version = version
        self.required_components = required_components
        self.app = app
        self.callback = callback

    def on_joined(self):
        self.app.session.call(
            'butler.register', self.application_name, self.version, self.required_components)
        if self.callback:
            self.callback()


class AutobahnButler:

    def __init__(self, application_name, version, required_components):
        self.app = None
        self._started = False
        self.application_name = application_name
        self.version = version
        self.required_components = required_components
        self._on_running_callbacks = []

    @property
    def session(self):
        return self.app.session

    def on_start_function(self):
        self._started = True
        for cb in self._on_running_callbacks:
            cb()
        self._on_running_callbacks = []

    def run_in_twisted(self, callback=None, url=DEFAULT_AUTOBAHN_ROUTER,
                       realm=DEFAULT_AUTOBAHN_REALM, blocking=False, **kwargs):
        self.app = AutobahnSync(self.application_name)
        app = self.app

        @app.subscribe('butler.' + self.application_name + '.start')
        def start_app():
            self.on_start_function()
        joined = OnJoined(
            self.application_name, self.version, self.required_components, self.app, self.callback)
        self.app.run_in_twisted(joined, url, realm, blocking, **kwargs)

    def run(self, url=DEFAULT_AUTOBAHN_ROUTER, realm=DEFAULT_AUTOBAHN_REALM,
            blocking=False, **kwargs):
        self.app = AutobahnSync(self.application_name)
        app = self.app

        @app.subscribe('butler.' + self.application_name + '.start')
        def start_app():
            self.on_start_function()
        joined = OnJoined(self.application_name, self.version, self.required_components, self.app)
        self.app.run(joined.on_joined, url, realm, blocking, **kwargs)

    def register(self, procedure=None, options=None):
        """Decorator for the :meth:`AutobahnButler.session.register`

        .. note::
            This decorator can be used before :meth:`AutobahnButler.run` is called.
            In such case the actual registration will be done at ``run()`` time.
        """

        def decorator(func):
            if self._started:
                self.app.session.register(endpoint=func, procedure=procedure, options=options)
            else:

                def registerer():
                    self.app.session.register(endpoint=func, procedure=procedure, options=options)

                # Wait for the WAMP session to be started
                self._on_running_callbacks.append(registerer)
            return func

        return decorator

    def subscribe(self, topic, options=None):
        """Decorator for the :meth:`AutobahnButler.session.subscribe`

        .. note::
            This decorator can be used before :meth:`AutobahnButler.run` is called.
            In such case the actual registration will be done at ``run()`` time.
        """

        def decorator(func):
            if self._started:
                self.app.session.subscribe(handler=func, topic=topic, options=options)
            else:

                def subscriber():
                    self.session.subscribe(handler=func, topic=topic, options=options)

                # Wait for the WAMP session to be started
                self._on_running_callbacks.append(subscriber)
            return func

        return decorator



from datetime import datetime
import mongoengine


class ConcurrencyError(Exception):
    pass


def _get_current_user():
    # TODO: need new implementation for crossbar
    return None
    # try:
    #     return current_user.id
    # except (RuntimeError, AttributeError):
    #     # If working outside flask context (i.g. init test for exemple)
    #     # there is no current user
    #     return None


class VersionedDocument(mongoengine.Document):

    """
    Mongoengine abstract document handling version, udpated and created fields
    as long as concurrent modifications handling
    """
    doc_version = mongoengine.IntField(required=True, default=1)
    doc_updated = mongoengine.DateTimeField(default=datetime.utcnow)
    doc_created = mongoengine.DateTimeField(default=datetime.utcnow)
    meta = {'abstract': True, '_version_bootstrapped': False}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bootstrap_version()

    @staticmethod
    def _version_pre_save(sender, document, **kwargs):
        """
        Update version and modified date on document change
        """
        if document.pk:
            document.doc_updated = datetime.utcnow()
            document.doc_version += 1

    @classmethod
    def _bootstrap_version(cls):
        if not cls._meta['_version_bootstrapped']:
            cls._meta['_version_bootstrapped'] = True
            # Signal to update metadata on document change
            mongoengine.signals.pre_save.connect(cls._version_pre_save, sender=cls)

    def save(self, *args, **kwargs):
        # Check for race condition on insert
        if self.pk:
            if 'save_condition' not in kwargs:
                kwargs['save_condition'] = {}
            if 'doc_version' not in kwargs['save_condition']:
                kwargs['save_condition']['doc_version'] = self.doc_version
        try:
            return super().save(*args, **kwargs)
        except mongoengine.errors.SaveConditionError:
            raise ConcurrencyError()


class HistorizedDocument(VersionedDocument):

    """
    Mongoengine abstract document handling history and race condition
    """
    meta = {'abstract': True, 'history_cls': None}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.history_cls = self._bootstrap_history_cls()
        # Shorthand to get the history of the current document
        self.get_history = lambda *args, **kwargs: \
            self.history_cls.objects(*args, origin=self, **kwargs).order_by('version')

    @classmethod
    def get_collection_history(cls):
        """Return history class for the document's collection"""
        return cls._bootstrap_history_cls()

    @staticmethod
    def _history_post_delete(sender, document):
        """Create HistoryItem on delete"""
        version = document.doc_version + 1
        history_cls = sender._meta['history_cls']
        item = history_cls(origin=document, author=_get_current_user(),
                           action='DELETE', version=version,
                           date=datetime.utcnow())
        item.save()

    @staticmethod
    def _history_post_save(sender, document, created):
        """Create HistoryItem document modification"""
        history_cls = sender._meta['history_cls']
        item = history_cls(origin=document,
                           author=_get_current_user(),
                           action='CREATE' if created else 'UPDATE',
                           # content=document.to_mongo().to_dict(),
                           content=document.to_json(),
                           version=document.doc_version,
                           date=document.doc_updated)
        item.save()

    @classmethod
    def _bootstrap_history_cls(cls):
        if cls._meta['history_cls']:
            return cls._meta['history_cls']
        collection = cls._meta['collection'] + '.history'
        assert collection != '.history', cls

        # Create history class with a dynamic name
        HistoryItem = type(cls.__name__ + 'History', (mongoengine.Document, ), {
            'meta': {'collection': collection},
            'origin': mongoengine.ReferenceField(cls, required=True),
            'author': mongoengine.ReferenceField('User'),
            # 'content': mongoengine.DictField(),
            'content': mongoengine.StringField(),
            'action': mongoengine.StringField(
                choices=['CREATE', 'UPDATE', 'DELETE'], required=True),
            'version': mongoengine.IntField(required=True),
            'date': mongoengine.DateTimeField(required=True),
        })

        cls._meta['history_cls'] = HistoryItem
        # Register signals to trigger history creation
        mongoengine.signals.post_save.connect(cls._history_post_save, sender=cls)
        mongoengine.signals.post_delete.connect(cls._history_post_delete, sender=cls)
        return cls._meta['history_cls']



from umongo import Document, Schema, fields
from .database import db


class Calendar(Document):

    class Schema(Schema):
        name = fields.StrField(required=True)
        users = fields.ListField(fields.EmailField(max_length=255), required=True)
        meetings = fields.ListField(fields.ReferenceField("Meeting"))

    class Config:
        collection = db.calendar



from ..model.calendar_model import Calendar
from ..view import (
    create_calendar, list_calendar, get_calendar, modify_calendar, event_calendar_creation)

from .common import BaseTest


class TestCalendarAPI(BaseTest):

    def test_create_calendar(self):
        self.clean_db()
        Calendar.collection.drop()
        r = create_calendar("test", ["test@test.com"])
        assert r
        calendar = Calendar.find_one(r.pk)
        assert calendar.pk == r.pk

    def test_list_calendars(self):
        self.clean_db()
        r = create_calendar("test", ["test@test.com"])
        r = create_calendar("test2", ["test@test.com"])
        r = create_calendar("test3", ["test@test.com"])
        r = create_calendar("test4", ["test@test.com"])
        r = list_calendar()
        assert r.count() == 4

    def test_get_calendars(self):
        r = list_calendar()
        assert r
        assert r.count() == 4
        cursor = r
        for elem in cursor:
            r = elem
            break
        r = get_calendar(r.pk)
        assert r
        assert r.data.name == 'test'

    def test_modify_calendars(self):
        r = create_calendar("test", ["test@test.com"])
        assert r
        r = modify_calendar(calendar_id=r.pk, name="new_name")
        assert r
        assert r.data.name == 'new_name'
        r = modify_calendar(calendar_id=r.pk, users=["pika@chu.com"])
        assert r
        assert len(r.data.users) == 1
        assert r.data.users[0] == "pika@chu.com"
        r = modify_calendar(calendar_id=r.pk, users=["tor@tank.com"], append=True)
        assert r
        assert len(r.data.users) == 2
        assert r.data.users[0] == "pika@chu.com"
        assert r.data.users[1] == "tor@tank.com"



class PaginationError(Exception):
    pass


def list_to_pagination(schema, data, page=1, per_page=20, **kwargs):
    total = kwargs.pop('total', len(data))
    return {
        '_items': data,
        '_meta': {
            'page': page,
            'per_page': per_page,
            'total': total
        }
    }


def paginate(schema, queryset, page=1, per_page=20):
    if isinstance(queryset, (list, tuple)):
        total = len(queryset)
    else:
        total = queryset.count()
    end = page * per_page
    start = end - per_page
    errors = []
    data = []
    for e in queryset[start:end]:
        res = schema.dump(e)
        if res.errors:
            errors.append(res.errors)
        data.append(res.data)
    if errors:
        raise PaginationError(errors)
    return {
        '_items': data,
        '_meta': {
            'page': page,
            'per_page': per_page,
            'total': total
        }
    }



from marshmallow import validates_schema, ValidationError
from marshmallow_mongoengine import ModelSchema, fields


# TODO: still useful with marshmallow>=2.6 ?
class UnknownCheckedSchema(ModelSchema):

    """
    ModelSchema with check for unknown field
    """

    @validates_schema(pass_original=True)
    def check_unknown_fields(self, data, original_data):
        for key in original_data:
            if key not in self.fields or self.fields[key].dump_only:
                raise ValidationError('Unknown field name {}'.format(key))


class BaseModelSchema(UnknownCheckedSchema):

    """
    Base schema to handle the default fields of a BaseModel
    """
    BASE_FIELDS = ('id', '_version', '_created', '_updated', '_links')
    id = fields.String(dump_only=True)
    _version = fields.Integer(dump_only=True, attribute="doc_version")
    _created = fields.DateTime(dump_only=True, attribute="doc_created")
    _updated = fields.DateTime(dump_only=True, attribute="doc_updated")
    # Shadow the model fields
    doc_version = fields.Skip(load_only=True, dump_only=True)
    doc_created = fields.Skip(load_only=True, dump_only=True)
    doc_updated = fields.Skip(load_only=True, dump_only=True)
    # TODO: move this to marshmallow-mongoengine ?
    _cls = fields.Skip(load_only=True, dump_only=True)



from ..model import Meeting, Calendar
from datetime import datetime

event_meeting_creation = "meeting.created"


def create_meeting(calendar_id, name, users=None, date_begin=None, duration=60,
                   place="Undefined", comment=None, tags=None):
    calendar = Calendar.find_one(calendar_id)
    if not calendar:
        return {'_errors': '404'}, 404

    if not users or not len(users):
        # todo call auth module to get the current user
        users = calendar.data.users

    if not date_begin:
        date_begin = datetime.utcnow()

    meeting = Meeting(calendar=calendar.pk, name=name, users=users,
                      tags=tags if tags else [], comment=comment if comment else "",
                      begin=date_begin, duration=duration, place=place)
    meeting.commit()
    return meeting


def list_meetings(calendar_id, page=1, per_page=20):
    skip = (page - 1) * per_page
    meetings = Meeting.find({'calendar': calendar_id}, skip=skip, limit=per_page)
    return meetings


def get_meeting(meeting_id):
    meeting = Meeting.find_one(meeting_id)
    return meeting


def modify_meeting(meeting_id, name=None, users=None, append=False, date_begin=None,
                   duration=None, place=None, comment=None, tags=None):
    meeting = Meeting.find_one(meeting_id)
    if not meeting:
        return None
    if name:
        meeting.data.name = name
    if users and len(users):
        if append:
            meeting.data.users.extend(users)
        else:
            meeting.data.users = users
    if date_begin:
        meeting.data.begin = date_begin
    if duration:
        meeting.data.duration = duration
    if place:
        meeting.data.place = place
    if comment:
        meeting.data.comment = comment
    if tags:
        meeting.data.tags = tags

    meeting.commit()
    return meeting


def delete_meeting(meeting_id):
    meeting = Meeting.find_one(meeting_id)
    if not meeting:
        return None
    # meeting.delete()
    # how to handle the delete via umongo?
    return meeting



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


REQUIRES = (
    "pysolr==3.4.0",
    "pymongo==2.9",
    "mongoengine==0.10.1",
    "marshmallow==2.6.0",
    "marshmallow-mongoengine==0.7.0",
    "Flask==0.10.1",
    "autobahn-sync",
    "umongo",
    "pyJWT",
    "passlib",
    "klein"
)


def read(fname):
    with open(fname) as fp:
        content = fp.read()
    return content


setup(
    name='xin-back',
    version='0.4.1',
    # description='',
    long_description=read('README.rst'),
    author='Scille',
    author_email='contact@scille.fr',
    url='https://github.com/Scille/xin-back',
    packages=find_packages(exclude=("test*", )),
    package_dir={'xin': 'xin'},
    include_package_data=True,
    install_requires=REQUIRES,
    license='MIT',
    zip_safe=False,
    # keywords='',
    classifiers=[
        'Intended Audience :: Developers',
        # 'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'
    ],
    test_suite='tests',
)



from marshmallow import ValidationError
# Republishing the default fields...
from marshmallow_mongoengine.fields import *  # noqa

from xin.bb.view_util.schema import UnknownCheckedSchema


# ...and add custom ones

# TODO: still useful with marshmallow>=2.6 ?
class UTCTZnaiveDateTime(DateTime):

    """
    DateTime object in UTC with no tzinfo
    """

    def _deserialize(self, value, attr, data):
        value = super()._deserialize(value, attr, data)
        if value and value.tzinfo:
            value = (value - value.utcoffset()).replace(tzinfo=None)
        return value


# TODO: still useful with marshmallow>=2.6 ?
class StrictBoolean(Boolean):

    """
    Real boolean (i.e. no cast from string) validation
    """

    def _deserialize(self, value, attr, data):
        if value not in (True, False, None):
            raise ValidationError('Not a boolean')
        return super()._deserialize(value, attr, data)


# TODO: still useful with marshmallow>=2.6 ?
class StrictString(String):

    """
    Real string (i.e. no desesperate cast) validation
    """

    def _deserialize(self, value, attr, data):
        if not isinstance(value, str):
            raise ValidationError('Not a string')
        return super()._deserialize(value, attr, data)


# TODO: still useful with marshmallow>=2.6 ?
class StrictList(List):

    """
    Real List (i.e. no cast from string) validation
    """

    def _deserialize(self, value, attr, data):
        if not isinstance(value, (list, tuple)):
            raise ValidationError('Not a list')
        # Replace default _deserialize instead of calling it to add
        # item index to error message
        errors = {}
        deserialized = []
        for i, each in enumerate(value):
            try:
                deserialized.append(self.container.deserialize(each))
            except ValidationError as err:
                errors[i] = err.messages
        if errors:
            raise ValidationError(errors)
        return deserialized


class LinkedReference(Reference):

    """
    Marshmallow custom field to map with :class Mongoengine.Reference:
    """

    def _deserialize(self, value, attr, data):
        # Incomming document can be 1) a single ObjectId,
        # 2) a dict containing an 'id' entry with the ObjectId
        if isinstance(value, dict):
            value = value.get('id')
            if not value:
                raise ValidationError("Champ 'id' manquant")
        return super()._deserialize(value, attr, data)

    def _serialize(self, value, attr, obj):
        # Return a dict containing the id and the links if registered for
        # this document type
        if value is None:
            return missing
        dump = {'id': str(getattr(value, 'pk', value.id))}
        if hasattr(value, 'get_links'):
            links = value.get_links()
            if links:
                dump['_links'] = links
        return dump


class LinkedGenericReference(GenericReference):

    """
    Marshmallow custom field to map with :class Mongoengine.GenericReference:
    """

    def _serialize(self, value, attr, obj):
        # Return a dict containing the id and the links if registered for
        # this document type
        if value is None:
            return missing
        dump = {'id': str(getattr(value, 'pk', value.id)),
                '_cls': value._class_name}
        if hasattr(value, 'get_links'):
            links = value.get_links()
            if links:
                dump['_links'] = links
        return dump


# Monkey patch to replace default DocumentReference field by our custom one
from marshmallow_mongoengine import register_field, register_field_builder  # noqa
from marshmallow_mongoengine.conversion.fields import (  # noqa
    ReferenceBuilder, ListBuilder, EmbeddedDocumentBuilder, GenericReferenceBuilder)

from xin.bb.model_util import fields as me_fields  # noqa


class ListBuilder(ListBuilder):
    MARSHMALLOW_FIELD_CLS = StrictList
register_field_builder(me_fields.ListField, ListBuilder)


class EmbeddedDocumentBuilder(EmbeddedDocumentBuilder):
    BASE_NESTED_SCHEMA_CLS = UnknownCheckedSchema
register_field_builder(me_fields.EmbeddedDocumentField, EmbeddedDocumentBuilder)

register_field(me_fields.BooleanField, StrictBoolean)
register_field(me_fields.StringField, StrictString)
register_field(me_fields.DateTimeField, UTCTZnaiveDateTime)
register_field(me_fields.ComplexDateTimeField, UTCTZnaiveDateTime)



import re
import mongoengine
from pysolr import SolrError
from flask import current_app

from xin.bb.tools import list_to_pagination
from xin.bb.model_util.version import VersionedDocument


class AsbFieldSolrConverter:

    """
    Abstract class for a converter from mongoengine field to solr

    A converter must implement two methods
    """
    FIELD_CLS = None
    FIELD_SOLR_EXTENSION = None

    def __init__(self, field_name, solr_field_name=None, extractor=None,
                 serializer=None, replace_aliases=None, aliases=None, multi=False):
        """
        :param field_name: name of the field
        :param solr_field_name: name of the field in solr
        :param extractor: function used to retrieve the field's
        value from it document
        :param serializer: function used to generate the dict of solr fields
        and their associated values
        :param replace_aliases: function used to replace the aliases in a query by
        their actual solr field names
        :param aliases: list of possible aliases of this field in a query
        :param multi: the extractor will returns a list of value, thus
        the solr field should be considered as multiple

        .. note : If multi=True the solr fields names will be corrected with a
            trailing "s" (e.g. "_s" ==> "_ss" for multi)
        """
        assert self.FIELD_CLS, 'FIELD_CLS must be set'
        assert self.FIELD_SOLR_EXTENSION, 'FIELD_SOLR_EXTENSION must be set'
        self._field_name = field_name
        if solr_field_name:
            self._solr_field_name = solr_field_name
        else:
            self._solr_field_name = field_name + self.FIELD_SOLR_EXTENSION
        if multi:
            self._solr_field_name += 's'
        self._extractor = extractor
        self._serializer = serializer
        self._replace_aliases = replace_aliases
        self._aliases = aliases or ()
        self._multi = multi

    def replace_aliases(self, query):
        """
        Look for the field possible aliases in the query and replace them by
        the actual solr field name
        """
        for alias in self._aliases:
            query = re.sub(r"\b%s\b" % alias, self._solr_field_name, query)
        if self._replace_aliases:
            return self._replace_aliases(query)
        else:
            return re.sub(r"\b%s\b" % self._field_name, self._solr_field_name, query)

    def _serialize(self, value):
        # Default serializer
        return {self._solr_field_name: value}

    def build_solr(self, doc):
        """
        Generate a dic of solr fields with their values from the given document
        """
        if self._extractor:
            value = self._extractor(doc)
        else:
            # Default value extractor
            value = getattr(doc, self._field_name, None)
        if not value:
            return {}
        if self._multi:
            if not isinstance(value, (list, tuple)):
                value = (value,)
            return self._multi_serialize_and_reduce(value)
        else:
            return self._serialize(value)

    def _multi_serialize_and_reduce(self, values):
        # Run the serialize for each value and merge the results
        reduced = {}
        for doc in [self._serialize(v) for v in values if v]:
            for key, value in doc.items():
                if key not in reduced:
                    reduced[key] = []
                reduced[key].append(value)
        return reduced

    @classmethod
    def can_convert(self, field_cls):
        """
        Determine if the given field can be handled by the converter
        """
        return issubclass(field_cls, self.FIELD_CLS)


class StringFieldSolrConverter(AsbFieldSolrConverter):
    FIELD_CLS = mongoengine.fields.StringField
    FIELD_SOLR_EXTENSION = '_s'


class IntFieldSolrConverter(AsbFieldSolrConverter):
    FIELD_CLS = mongoengine.fields.IntField
    FIELD_SOLR_EXTENSION = '_i'


class FloatFieldSolrConverter(AsbFieldSolrConverter):
    FIELD_CLS = mongoengine.fields.FloatField
    FIELD_SOLR_EXTENSION = '_f'


class DateTimeFieldSolrConverter(AsbFieldSolrConverter):
    FIELD_CLS = mongoengine.fields.DateTimeField
    FIELD_SOLR_EXTENSION = '_dt'


class BooleanFieldSolrConverter(AsbFieldSolrConverter):
    FIELD_CLS = mongoengine.fields.BooleanField
    FIELD_SOLR_EXTENSION = '_b'


class GenericReferenceFieldSolrConverter(AsbFieldSolrConverter):
    FIELD_CLS = mongoengine.fields.GenericReferenceField
    FIELD_SOLR_EXTENSION = '_r'

    def _serialize(self, value):
        return {self._solr_field_name: str(value.id)}


class ReferenceFieldSolrConverter(AsbFieldSolrConverter):
    FIELD_CLS = mongoengine.fields.ReferenceField
    FIELD_SOLR_EXTENSION = '_r'

    def _serialize(self, value):
        return {self._solr_field_name: str(value.id)}


class SolrFieldConverterManager:

    """
    Store the available converters class and provide a convenient way
    to tests
    """

    def __init__(self):
        # Default converters for base types
        self._converters = [
            StringFieldSolrConverter,
            IntFieldSolrConverter,
            FloatFieldSolrConverter,
            DateTimeFieldSolrConverter,
            BooleanFieldSolrConverter,
            GenericReferenceFieldSolrConverter,
            ReferenceFieldSolrConverter
        ]

    def register_converter_cls(self, converter_cls):
        """
        Add a new converter class.
        """
        # Register backward to prevent base generic fields to shadow
        self._converters.insert(0, converter_cls)

    def build_converter(self, field_name, field, **kwargs):
        """
        Retrieve and build a converter for the given field

        :param field_name: name of the field
        :param field: field on which the converter will be used
        :param kwargs: additional params passed to converter ``__init__``

        .. note : If multiple converter can match the field, the one
            added last will be used
        """
        if isinstance(field, mongoengine.fields.ListField):
            kwargs.setdefault('multi', True)
            field_cls = type(field.field)
        else:
            kwargs.setdefault('multi', False)
            field_cls = type(field)
        converter = next((c for c in self._converters
                          if c.can_convert(field_cls)), None)
        return converter(field_name, **kwargs)


# Created a single default instance and register shortcuts
solr_field_converter_manager = SolrFieldConverterManager()
solr_register_field_converter = solr_field_converter_manager.register_converter_cls
solr_build_converter = solr_field_converter_manager.build_converter


class Searcher:

    """
    Abstract default searcher class
    """

    def __init__(self, document_cls):
        self.document_cls = document_cls

    def search_or_abort(self, **kwargs):
        raise NotImplementedError()

    def build_document(self, document):
        """Register into solr the current mongoengine document"""
        raise NotImplementedError()

    def clear_document(self, document):
        """Remove from solr the current mongoengine document"""
        document = document or self.document
        raise NotImplementedError()

    @staticmethod
    def on_post_delete(sender, document):
        document.searcher.clear_document(document)

    @staticmethod
    def on_post_save(sender, document, created):
        document.searcher.build_document(document)


def get_document_base_type(document):
    return document._class_name.split('.')[0]


class BaseSolrSearcher(Searcher):

    """
    Base searcher class for solr integration
    """

    FIELDS = ()

    def __init__(self, *args, converters=None, **kwargs):
        super().__init__(*args, **kwargs)
        dcls = self.document_cls
        # Add default fields
        self.converters = [
            StringFieldSolrConverter('id', extractor=lambda doc: str(doc.id),
                                     solr_field_name='doc_id'),
            StringFieldSolrConverter('_class_name', solr_field_name='doc_type',
                                     replace_aliases=lambda x: x),
            StringFieldSolrConverter('_class_name', solr_field_name='doc_base_type',
                                     extractor=get_document_base_type, replace_aliases=lambda x: x),
            IntFieldSolrConverter('doc_version', aliases=('_version',)),
            DateTimeFieldSolrConverter('doc_updated', aliases=('_updated',)),
            DateTimeFieldSolrConverter('doc_created', aliases=('_created',)),
        ]
        for field in self.FIELDS:
            field_cls = getattr(dcls, field, None)
            if field_cls:
                self.converters.append(solr_build_converter(field, field_cls))

    def build_and_register_converter(self, *args, **kwargs):
        self.converters.append(solr_build_converter(*args, **kwargs))

    def register_custom_field(self, field_name, field_cls, field_extractor, alias=None):
        """
        Register a special field for the solr document

        :param field_name: name of the field in solr
        :param field_cls: mongoengine class of the field
        :param field_extractor: function that return the field's value
        from a given document or None
        :param alias: list of additional alias for the field
        """
        self.converters.append(
            solr_build_converter(field_name, field_cls, extractor=field_extractor, aliases=alias))

    def _replace_aliases(self, query):
        for c in self.converters:
            query = c.replace_aliases(query)
        return query

    def search_or_abort(self, q=None, fq=None, sort=None, page=1, per_page=20):
        """

        :param q: solr style query
        :param fq: solr style list of query filters
        :param sort: solr style list of query sorters
        :param page: current page to retrieve
        :param per_page: number of elements per page
        :return: a ``Pagination`` of mongoengine documents
        """
        if not q:
            q = '*:*'
        else:
            q = self._replace_aliases(q)
        kwargs = {'start': (page - 1) * per_page, 'rows': per_page, 'q': q}
        if fq:
            kwargs['fq'] = [self._replace_aliases(e) for e in fq]
        if sort:
            kwargs['sort'] = [self._replace_aliases(e) for e in sort]
        try:
            docs = self.solr_search(**kwargs)
            hits = docs.hits
        except SolrError as e:
            current_app.logger.warning('SolrError: %s' % str(e))
            docs = ()
            hits = 0
        # Pagination is already handled by solr, thus we use a dummy
        # one in mongo and correct it according to solr's query
        items = self.document_cls.objects(pk__in=[d['doc_id'] for d in docs])
        if 'sort' in kwargs:
            # If needed, reorder the results from mongo to match the original
            # order provided by solr
            ordered_items = []
            for d in docs:
                doc_id = d['doc_id']
                for item in items:
                    pk = str(item.pk)
                    if pk == doc_id:
                        ordered_items.append(item)
            items = ordered_items
        items = list_to_pagination(items, already_sliced=True, page=page,
                                   per_page=per_page, total=hits,
                                   q=q, fq=fq, sort=sort)
        return items

    def generate_solr_doc(self, document):
        """
        Create a solr document from a mongoengine document
        """
        sdoc = {
            'id': '%s-%s' % (document._class_name, document.pk),
        }
        for c in self.converters:
            sdoc.update(c.build_solr(document))
        return sdoc

    def build_document(self, document, **kwargs):
        """
        Register into solr the current mongoengine document
        """
        kwargs.setdefault('commit', False)
        kwargs.setdefault('waitFlush', False)
        doc = self.generate_solr_doc(document)
        # Given we use document's pk as solr id, no need to clear the
        # previous solr document (will be replace by the new one)
        current_app.solr.add((doc,), **kwargs)

    def clear_document(self, document, **kwargs):
        """
        Remove from solr the current mongoengine document
        """
        kwargs.setdefault('commit', False)
        kwargs.setdefault('waitFlush', False)
        if document.pk:
            sdoc_id = '%s-%s' % (document._class_name, document.pk)
            current_app.solr.delete(id=sdoc_id, **kwargs)

    def clear_collection(self, **kwargs):
        """
        Remove the entire collection from solr
        """
        kwargs.setdefault('commit', False)
        kwargs.setdefault('waitFlush', False)
        base_type = get_document_base_type(self.document_cls)
        current_app.solr.delete(q='doc_base_type:' + base_type, **kwargs)

    def solr_search(self, q, fq=None, sort=None, **kwargs):
        """
        Do a solr query limited to the current collection
        """
        fq = fq or []
        if sort:
            if isinstance(sort, (list, tuple)):
                sort = ','.join(sort)
            kwargs['sort'] = sort
        base_type = get_document_base_type(self.document_cls)
        fq.append('doc_base_type:' + base_type)
        return current_app.solr.search(q, fq=fq, **kwargs)


class SearchableDocument(VersionedDocument):

    """
    Mongoengine abstract document providing search functionalities
    """
    meta = {'abstract': True, 'searcher_cls': None, '_searcher': None}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        searcher_cls = self._meta.get('searcher_cls')
        if searcher_cls:
            self.searcher = self._search_bootstrap()

    @classmethod
    def _search_bootstrap(cls):
        if cls._meta['_searcher']:
            return cls._meta['_searcher']
        # Register signals to trigger solr sync
        searcher_cls = cls._meta.get('searcher_cls')
        if not searcher_cls:
            raise NotImplementedError('No searcher setted for this document')
        searcher = searcher_cls(cls)
        mongoengine.signals.post_save.connect(searcher.on_post_save, sender=cls)
        mongoengine.signals.post_delete.connect(searcher.on_post_delete, sender=cls)
        cls._meta['_searcher'] = searcher
        return searcher

    @classmethod
    def search_or_abort(cls, *args, **kwargs):
        """
        Shortcut to ``SearchableDocument.searcher.search_or_abort``
        """
        return cls._search_bootstrap().search_or_abort(*args, **kwargs)



from os import environ


SECRET_KEY = environ.get('SECRET_KEY', 'secret_for_test_only!')
REGISTER_USER_RPC = environ['AUTHENTIC_REGISTER_USER_RPC']
RETRIEVE_USER_RPC = environ['AUTHENTIC_RETRIEVE_USER_RPC']
LISTEN_PORT = environ.get('AUTHENTIC_LISTEN_PORT', 8081)
TOKEN_VALIDITY = environ.get('AUTHENTIC_TOKEN_VALIDITY', 7 * 24 * 3600)
CORS_ORIGIN = environ.get('AUTHENTIC_CORS_ORIGIN', '*').encode()



from ..model.calendar_model import Calendar
from ..model.meeting import Meeting


class BaseTest:

    @classmethod
    def clean_db(cls):
        Calendar.collection.drop()
        Meeting.collection.drop()
